//		Copyright © 1998	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: StOpenClose.cpp,v $
//		Revision 1.8  1999/01/09 15:33:39  jjk
//		Explicitly include the prefix header file (R4 x86 support).
//
//		Revision 1.7  1998/10/23 09:25:15  jjk
//		Add support for multiple partitions per physical device.
//
//		Revision 1.6  1998/06/21 10:03:27  jjk
//		Backspace over second EOF at EOD when doing MTIOCTOP MTEOD.
//
//		Revision 1.5  1998/06/12 09:52:13  jjk
//		Allow second UNIT ATTENTION after loading medium.
//		Encapsulate debugging messages.
//
//		Revision 1.4  1998/06/11 22:13:34  jjk
//		Locking semaphore is now permanently stored in the device descriptor.
//
//		Revision 1.3  1998/06/11 16:12:32  jjk
//		Load medium after receiving a unit attention condition.
//
//		Revision 1.2  1998/06/11 10:16:25  jjk
//		Allow open() to succeed even if drive is not ready.
//		Try to bring drive on-line before each subsequent operation.
//
//		Revision 1.1  1998/06/06 15:20:47  jjk
//		Initial revision
//

//		Implementation of driver operations open(), close(), free().

#include "Prefix.h"

#include <stdio.h>
#include <string.h>

#include <CAM.h>
#include <KernelExport.h>

#include "mtio.h"
#include "SCSICommands.h"
#include "SCSIData.h"
#include "StDevices.h"
#include "StOpenClose.h"

//------------------------------------------------------------------------------
// Find block size limits of the given device.
static status_t
scsiReadBlockLimits(PhysicalDevice &device)
{
	CDBReadBlockLimits cdb(device.mLun);
	SCSIReadBlockLimitsData blockLimits;
	status_t status =
		scsiCommand(device, CAM_DIR_IN, cdb, &blockLimits, sizeof(blockLimits),
						5, kRetries);
	if (status == B_OK) {
		device.mMinBlockSize = blockLimits.minBlockLength();
		device.mMaxBlockSize = blockLimits.maxBlockLength();
	}
	
	return status;
}

// Determine the current tape position.
static status_t
scsiReadPosition(PhysicalDevice &device)
{
	CDBReadPosition cdb(device.mLun);
	SCSIReadPositionData position;
	status_t status =
		scsiCommand(device, CAM_DIR_IN, cdb, &position, sizeof(position),
						5, kRetries);
	if (status == B_OK) {
		device.mBOT = position.bop;
		device.mEOT = position.eop;
	}
	
	return status;
}

// Configure the device.
static status_t
configureDataCompression(PhysicalDevice &device, bool compression)
{
	struct {
		SCSIModeParameterHeader6 h;
		SCSIDataCompressionPage p;
	} curParList, chgParList;
	
	// Determine current configuration.
	if (scsiModeSense(device, SCSIDataCompressionPage::kPageCode,
							&curParList, sizeof(curParList), true) < B_OK) {
		// Assume that the device doesn't implement this page.
		device.mDataCompression = false;
		return B_OK;
	}
	device.mDataCompression = compression;

	// Find out which of these values are changeable.
	status_t status = scsiModeSense(device, SCSIDataCompressionPage::kPageCode,
											  &chgParList, sizeof(chgParList), true,
											  CDBModeSense0::kChangeableValues);
	if (status < B_OK) {
		return status;
	}

	if (compression) {
		// Enable data compression.
		curParList.p.dce |= chgParList.p.dce;
		if (chgParList.p.compressionAlgorithm) {
			curParList.p.compressionAlgorithm =
				SCSIDataCompressionPage::kDefaultCompression;
		}
	} else {
		// Disable data compression.
		curParList.p.dce &= ~(chgParList.p.dce);
	}

	// Always enable decompression, if possible.
	curParList.p.dde |= chgParList.p.dde;
	curParList.p.red &= ~(chgParList.p.red);
	if (chgParList.p.decompressionAlgorithm) {
		curParList.p.decompressionAlgorithm =
				SCSIDataCompressionPage::kDefaultCompression;
	}
	
	curParList.h.modeDataLength = 0;				// reserved for MODE SELECT
		
	status = scsiModeSelect(device, &curParList, sizeof(curParList));
	if (status < B_OK) {
		ddprintf(("ST - MODE SELECT 0x0F failed on device %d/%d/%d.\n",
					 device.mPathID, device.mTargetID, device.mLun));
		return status;
	}

	return status;
}

static status_t
configureDevice(PhysicalDevice &device, int partition, bool compression)
{
	struct {
		SCSIModeParameterHeader6 h;
		SCSIDeviceConfigurationPage p;
	} curParList, chgParList;
	
	// Determine current device configuration.
	status_t status =
		scsiModeSense(device, SCSIDeviceConfigurationPage::kPageCode,
						  &curParList, sizeof(curParList), true);
	if (status < B_OK) {
		return status;
	}
	device.mWriteProtected = curParList.h.tape.wp;

	// Find out which of these values are changeable.
	status = scsiModeSense(device, SCSIDeviceConfigurationPage::kPageCode,
								  &chgParList, sizeof(chgParList), true,
								  CDBModeSense0::kChangeableValues);
	if (status < B_OK) {
		return status;
	}

	// Reconfigure the device.
	if (partition != curParList.p.activePartition) {
		// Try to change the active partition.
		ddprintf(("ST - Change active partition (%d -> %d).\n",
					 curParList.p.activePartition, partition));
		if (chgParList.p.cap && chgParList.p.activePartition) {
			curParList.p.cap = true;
			curParList.p.activePartition = partition;
		} else {
			ddprintf(("ST - Cannot change partition.\n"));
			return B_BAD_VALUE;
		}
	}
	curParList.p.rsmk |= chgParList.p.rsmk;	// report setmarks
	curParList.p.rew	&= ~(chgParList.p.rew);	// don't report early warning
	curParList.p.sew	|= chgParList.p.sew;		// synchronize at early warning
	if (chgParList.p.dataCompression) {
		// This device selects data compression as specified in the standard.
		device.mDataCompression = compression;
		curParList.p.dataCompression =
			compression
			? SCSIDeviceConfigurationPage::kDefaultCompression
			: SCSIDeviceConfigurationPage::kNoCompression;
	}

	curParList.h.modeDataLength = 0;				// reserved for MODE SELECT

	status = scsiModeSelect(device, &curParList, sizeof(curParList));
	if (status < B_OK) {
		ddprintf(("ST - MODE SELECT 0x10 failed on device %d/%d/%d.\n",
					 device.mPathID, device.mTargetID, device.mLun));
		return status;
	}

	if (!(chgParList.p.dataCompression)) {
		// This device doesn't select data compression via page 0x10.
		// Try the alternative "Data Compression Page", 0xf.
		status = configureDataCompression(device, compression);
	}

	return status;
}

//------------------------------------------------------------------------------
// Try to bring the device on-line.
static bool isQIC(const PhysicalDevice &device)
{
	// Determine if this is a QIC device.
	switch (device.mDensity) {

	case kDensityQIC11:
	case kDensityQIC24:
	case kDensityQIC120:
	case kDensityQIC150:
	case kDensityQIC525:
	case kDensityQIC1350:
	case kDensityQIC2100:
	case kDensityQIC2GB:
	case kDensityQIC5GB:
		return true;
	}
	
	return false;
}

status_t
mountDevice(LogicalDevice &logicalDevice)
{
	PhysicalDevice &device = logicalDevice.mPhys;
	if (device.mOnline) {
		// Device is already on-line.
		return B_OK;
	}
	
	// Check if device is ready.
	bool gotUnitAttention;
	status_t status = deviceReady(device, gotUnitAttention);
	if (status < B_OK) {
		goto err0;
	}

	// If a unit attention condition was encountered, load the medium.
	// This may cause another unit attention.
	if (gotUnitAttention) {
		ddprintf(("ST - UNIT ATTENTION received, load medium.\n"));
		if ((status = scsiLoadUnload(device, true)) < B_OK
			 || (status = deviceReady(device, gotUnitAttention)) < B_OK) {
			goto err0;
		}
	}
	
	// Prevent removal of the medium, if applicable.
	if (device.mRemovableMedium) {
		(void)scsiPreventMediumRemoval(device, true);
	}
	
	// Get the device's block size limits and determine the block size.
	if ((status = scsiReadBlockLimits(device)) < B_OK) {
		goto err1;
	}
	if (device.mMinBlockSize == device.mMaxBlockSize) {
		// Device supports blocks of one size only.
		device.mBlockSize = device.mMinBlockSize;
	}
	// Try to set the block size.
	(void)setBlockSize(device, device.mBlockSize);

	// Get the resulting block size and the density.
	status = getBlockSize(device, device.mBlockSize, device.mDensity);
	if (status < B_OK) {
		goto err1;
	}
	
	// Configure the device.
	status = configureDevice(device,
									 logicalDevice.mPartition,
									 logicalDevice.mCompression);
	if (status < B_OK) {
		goto err1;
	}

	// Get the current tape position.
	if ((status = scsiReadPosition(device)) < B_OK) {
		goto err1;
	}
	
	// Check write protection.
	if (device.mWriteProtected && device.mWriteOK) {
		status = B_PERMISSION_DENIED;
		goto err1;
	}
	
	// Initialize device.
	device.mWrittenTo = false;
	device.mWriteEOF = true;
	device.mWrite2EOF = !isQIC(device);
	device.mWrittenPastEOM = false;
	device.mWriteBarrier = true;
	device.mStatusPending = false;
	device.mOnline = true;

	return B_OK;
	
	// Error exits - restore what we changed and return error status.
err1:
	if (device.mRemovableMedium) {
		(void)scsiPreventMediumRemoval(device, false);
	}
err0:
	return status;
}

//------------------------------------------------------------------------------
// Open a tape device.
status_t
stOpen(const char *pName, uint32 flags, void **ppCookie)
{
	// Find the logical device to be opened.
	LogicalDevice *pLogicalDevice = findDevice(pName);
	if (!pLogicalDevice) {
		// Invalid device name.
		ddprintf(("ST - Invalid name '%s'.\n", pName));
		return B_BAD_VALUE;
	}
	*ppCookie = pLogicalDevice;

	// Get the corresponding physical device.
	PhysicalDevice &device = pLogicalDevice->mPhys;
		
	// Obtain exclusive access to the physical device.
	if (atomic_or(&(device.mFlags), PhysicalDevice::kInUse)) {
		// Device already in use.
		ddprintf(("ST - Device %d/%d/%d already in use.\n",
					 device.mPathID, device.mTargetID, device.mLun));
		return B_BUSY;
	}
	
	// Evaluate flags.
	status_t status = B_OK;
	switch (flags & O_RWMASK) {
	
	case O_RDONLY:
		device.mReadOK = true;
		device.mWriteOK = false;
		break;
	
	case O_WRONLY:
		device.mReadOK = false;
		device.mWriteOK = true;
		break;
	
	case O_RDWR:
		device.mReadOK = true;
		device.mWriteOK = true;
		break;
	
	default:
		status = B_BAD_VALUE;
		goto err0;
	}
	
	// Note that device has been opened and try to bring it on-line.
	atomic_or(&(device.mFlags), PhysicalDevice::kOpen);
	device.mOnline = false;
	(void)mountDevice(*pLogicalDevice);

	return status;

	// Error exit.
err0:
	atomic_and(&(device.mFlags), ~PhysicalDevice::kInUse);
	return status;
}

//------------------------------------------------------------------------------
// Close a tape device.
status_t
stClose(void *pCookie)
{
	LogicalDevice *pLogicalDevice = (LogicalDevice *)pCookie;
	PhysicalDevice &device = pLogicalDevice->mPhys;
	DeviceLock lock(device);
		
	// Disable all further operations.
	atomic_and(&(device.mFlags), ~PhysicalDevice::kOpen);
		
	return B_OK;
}

//------------------------------------------------------------------------------
// Free tape device after I/O has completed.
status_t
stFree(void *pCookie)
{
	LogicalDevice *pLogicalDevice = (LogicalDevice *)pCookie;
	PhysicalDevice &device = pLogicalDevice->mPhys;
	DeviceLock lock(device);
	
	status_t status = B_OK;
	if (device.mOnline) {
		// Perform final rites on the tape.  This is much too complicated,
		// but I'm working to a specification here...	
		if (device.mWrittenTo && device.mWriteEOF) {
			ddprintf(("ST - Device written to, trying to write EOF.\n"));
			// Write one filemark for QIC devices, two filemarks otherwise.
			const uint32 nMarks = device.mWrite2EOF ? 2 : 1;
			if ((status = scsiWriteFilemarks(device, nMarks)) < B_OK) {
				// Tolerate an EOM condition.
				if (status != kStatusNotComplete || !device.mEOM) {
					goto err;
				}
			}
			
			if (pLogicalDevice->mAutoRewind) {
				ddprintf(("ST - Rewind written-to device.\n"));
				// Rewind to BOT.
				if ((status = scsiRewind(device, true)) < B_OK) {
					goto err;
				}
			
			} else if (device.mWrite2EOF) {
				ddprintf(("ST - Backward space written-to device.\n"));
				// Move back over the second filemark (not for QIC devices).
				if ((status = scsiSpace(device, CDBSpace::kFilemarks, -1)) < B_OK) {
					goto err;
				}
			}
		
		} else if (pLogicalDevice->mAutoRewind) {
			ddprintf(("ST - Rewind device.\n"));
			// Rewind.
			if ((status = scsiRewind(device, true)) < B_OK) {
				goto err;
			}
	
		} else if (!pLogicalDevice->mBSDBehavior) {
			ddprintf(("ST - AT&T no-auto-rewind device.\n"));
			// Forward space past next filemark, unless already done.
			if (!(device.mFilemark||device.mSetmark||device.mEOM||device.mEOD)) {
				ddprintf(("ST - Forward space past filemark.\n"));
				if ((status = scsiSpace(device, CDBSpace::kFilemarks, 1)) < B_OK) {
					goto err;
				}
			}
			
		} else {
			ddprintf(("ST - Berkeley no-auto-rewind device.\n"));
			// Leave it as it is.
		}
	}

err:	
	// Allow medium removal (if applicable).
	if (device.mRemovableMedium) {
		(void)scsiPreventMediumRemoval(device, false);
	}

	// Relinquish our hold on the device.
	atomic_and(&(device.mFlags), ~PhysicalDevice::kInUse);

	return status;
}

static char RCSid[] = "$Id: StOpenClose.cpp,v 1.8 1999/01/09 15:33:39 jjk Exp $";
